@@ -1,7 +1,7 @@
 from .utils import appdir
 import g4f
 
-__version__ = "0.4.2"
+__version__ = "0.4.7"
 __author__ = "Smartwa"
 __repo__ = "https://github.com/Simatwa/python-tgpt"
 
@@ -14,6 +14,9 @@ tgpt_providers = [
     "phind",
     "llama2",
     "blackboxai",
+    "gpt4all",
+    "webchatgpt",
+    "g4fauto",
 ]
 
 gpt4free_providers = [
