@@ -59,6 +59,8 @@ class this:
 
     getExc = lambda e: e.args[1] if len(e.args) > 1 else str(e)
 
+    context_settings = dict(auto_envvar_prefix="PYTGPT")
+
     """Console utils"""
 
     @staticmethod
@@ -332,7 +334,7 @@ class Main(cmd.Cmd):
         chat_completion=False,
         ignore_working=False,
         rawdog=False,
-        external_exec=False,
+        internal_exec=False,
         confirm_script=False,
         interpreter="python",
         *args,
@@ -352,7 +354,7 @@ class Main(cmd.Cmd):
 
                 self.RawDog = RawDog(
                     quiet=quiet,
-                    external_exec=external_exec,
+                    internal_exec=internal_exec,
                     confirm_script=confirm_script,
                     interpreter=interpreter,
                     prettify=True,
@@ -360,7 +362,37 @@ class Main(cmd.Cmd):
                 intro = self.RawDog.intro_prompt
                 getpass.getuser = lambda: "RawDog"
 
-            if provider == "leo":
+            if provider == "g4fauto":
+                from pytgpt.gpt4free.utils import TestProviders
+
+                test = TestProviders(quiet=quiet, timeout=timeout)
+                g4fauto = test.best if ignore_working else test.auto
+                if isinstance(g4fauto, str):
+                    provider = "g4fauto+" + g4fauto
+                    from pytgpt.gpt4free import GPT4FREE
+
+                    self.bot = GPT4FREE(
+                        provider=g4fauto,
+                        auth=auth,
+                        max_tokens=max_tokens,
+                        model=model,
+                        chat_completion=chat_completion,
+                        ignore_working=ignore_working,
+                        timeout=timeout,
+                        intro=intro,
+                        filepath=filepath,
+                        update_file=update_file,
+                        proxies=proxies,
+                        history_offset=history_offset,
+                        act=awesome_prompt,
+                    )
+                else:
+                    raise Exception(
+                        "No working g4f provider found. "
+                        "Consider running 'pytgpt gpt4free test -y' first"
+                    )
+
+            elif provider == "leo":
                 import pytgpt.leo as leo
 
                 self.bot = leo.LEO(
@@ -437,9 +469,8 @@ class Main(cmd.Cmd):
             elif provider == "bard":
                 from pytgpt.bard import BARD
 
-                assert (
-                    auth
-                ), f"Path to bard.google.com.cookie.json file is required. Use the flag `--key` or `-k`"
+                assert auth, f"Path to bard.google.com.cookie.json file is required. "
+                "Use the flag `--key` or `-k`"
                 self.bot = BARD(
                     cookie_file=auth,
                     proxy=proxies,
@@ -460,6 +491,7 @@ class Main(cmd.Cmd):
                     history_offset=history_offset,
                     act=awesome_prompt,
                     model=getOr(model, phind.default_model),
+                    quiet=quiet,
                 )
 
             elif provider == "llama2":
@@ -498,6 +530,42 @@ class Main(cmd.Cmd):
                     act=awesome_prompt,
                 )
 
+            elif provider == "gpt4all":
+                assert auth, f"Path to LLM (.gguf or .bin) file is required. "
+                "Download models from <https://gpt4all.io>. "
+                "Use the flag `--key` or `-k`"
+
+                from pytgpt.gpt4all import GPT4ALL
+
+                self.bot = GPT4ALL(
+                    model=auth,
+                    is_conversation=disable_conversation,
+                    max_tokens=max_tokens,
+                    temperature=temperature,
+                    presence_penalty=top_p,
+                    frequency_penalty=top_k,
+                    top_p=top_p,
+                    intro=intro,
+                    filepath=filepath,
+                    update_file=update_file,
+                    history_offset=history_offset,
+                )
+
+            elif provider == "webchatgpt":
+                assert auth, f"Path to `chat.openai.cookies.com.json` is required. "
+                "Use the flag `--key` or `-k`"
+
+                from pytgpt.webchatgpt import main
+
+                self.bot = main.WEBCHATGPT(
+                    cookie_file=auth,
+                    model=getOr(model, main.default_model),
+                    proxy=proxies,
+                    timeout=timeout,
+                    filepath=filepath,
+                    update_file=update_file,
+                )
+
             elif provider in pytgpt.gpt4free_providers:
                 from pytgpt.gpt4free import GPT4FREE
 
@@ -535,6 +603,9 @@ class Main(cmd.Cmd):
         self.disable_stream = False
         self.provider = provider
         self.disable_coloring = False
+        self.internal_exec = internal_exec
+        self.confirm_script = confirm_script
+        self.interpreter = interpreter
         self.rawdog = rawdog
         self.__init_time = time.time()
         self.__start_time = time.time()
@@ -781,6 +852,15 @@ class Main(cmd.Cmd):
         if click.confirm("Do you wish to run the command(s) generated in your system"):
             self.do_sys(self.bot.get_message(self.bot.last_response))
 
+    @busy_bar.run("While changing directory")
+    def do_cd(self, line):
+        """Change directory
+        Usage :
+             cd <path-to-directory>
+        """
+        assert line, "File path is required"
+        os.chdir(line)
+
     def do_clear(self, line):
         """Clear console"""
         sys.stdout.write("\u001b[2J\u001b[H")
@@ -833,6 +913,30 @@ class Main(cmd.Cmd):
             is_json=True,
         )
 
+    @busy_bar.run()
+    def do_exec(self, line):
+        """Exec python code in last response with RawDog"""
+        last_response = self.bot.get_message(self.bot.last_response)
+        assert last_response, "Last response is null"
+        assert "```python" in last_response, "Last response has no python code"
+        if self.rawdog:
+            self.RawDog.main(last_response)
+        else:
+            rawdog = RawDog(
+                quiet=self.quiet,
+                internal_exec=self.internal_exec,
+                confirm_script=self.confirm_script,
+                interpreter=self.interpreter,
+                prettify=self.prettify,
+            )
+            rawdog.main(last_response)
+
+    @busy_bar.run()
+    def do_rawdog(self, line):
+        """Repeat executing last rawdog's python code"""
+        assert self.rawdog, "Session not in rawdog mode. Restart with --rawdog"
+        self.default(self.bot.get_message(self.bot.last_response))
+
     @busy_bar.run()
     def default(self, line, exit_on_error: bool = False):
         """Chat with LLM"""
@@ -845,10 +949,11 @@ class Main(cmd.Cmd):
             self.__start_time = time.time()
             busy_bar.start_spinning()
             ai_response = self.bot.chat(line, stream=False)
+            busy_bar.stop_spinning()
             is_feedback = self.RawDog.main(ai_response)
             if is_feedback:
                 return self.default(self.bot.chat(is_feedback))
-            busy_bar.stop_spinning()
+            self.__end_time = time.time()
 
         else:
             self.__start_time = time.time()
@@ -962,7 +1067,7 @@ class ChatInteractive:
     """Interactive command"""
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.option(
         "-m",
         "--model",
@@ -999,8 +1104,7 @@ class ChatInteractive:
     @click.option(
         "-k",
         "--key",
-        envvar="auth_key",
-        help="LLM API access key or auth value",
+        help="LLM API access key or auth value or path to LLM with provider.",
     )
     @click.option(
         "-ct",
@@ -1053,7 +1157,6 @@ class ChatInteractive:
     @click.option(
         "-ho",
         "--history-offset",
-        envvar="history_offset",
         help="Limit conversation history to this number of last texts",
         type=click.IntRange(100, 16000),
         default=10250,
@@ -1071,7 +1174,6 @@ class ChatInteractive:
         "-pp",
         "--proxy-path",
         type=click.Path(exists=True),
-        envvar="proxy_path",
         help="Path to .json file containing proxies",
     )
     @click.option(
@@ -1080,7 +1182,6 @@ class ChatInteractive:
         type=click.Choice(pytgpt.available_providers),
         default=this.default_provider,
         help="Name of LLM provider.",
-        envvar="llm_provider",
         metavar=(
             f"[{', '.join(pytgpt.tgpt_providers)}] etc, "
             "run 'pytgpt gpt4free list providers -w' to view more providers"
@@ -1089,7 +1190,6 @@ class ChatInteractive:
     @click.option(
         "-vo",
         "--vertical-overflow",
-        envvar="vertical_overflow",
         help="Vertical overflow behaviour on content display",
         type=click.Choice(["visible", "crop", "ellipsis"]),
         default="ellipsis",
@@ -1105,7 +1205,7 @@ class ChatInteractive:
         "-q",
         "--quiet",
         is_flag=True,
-        help="Flag for controlling response-framing",
+        help="Flag for controlling response-framing and response verbosity",
         default=False,
     )
     @click.option(
@@ -1142,10 +1242,10 @@ class ChatInteractive:
         help="Generate and auto-execute Python scripts - (experimental)",
     )
     @click.option(
-        "-ex",
-        "--external-exec",
+        "-ix",
+        "--internal-exec",
         is_flag=True,
-        help="RawDog : Execute scripts with system's python interpreter",
+        help="RawDog : Execute scripts with exec function instead of out-of-script interpreter",
     )
     @click.option(
         "-cs",
@@ -1158,7 +1258,6 @@ class ChatInteractive:
         "--interpreter",
         default="python",
         help="RawDog : Python's interpreter name",
-        envvar="python_interpreter",
     )
     @click.help_option("-h", "--help")
     def interactive(
@@ -1191,7 +1290,7 @@ class ChatInteractive:
         chat_completion,
         ignore_working,
         rawdog,
-        external_exec,
+        internal_exec,
         confirm_script,
         interpreter,
     ):
@@ -1217,7 +1316,7 @@ class ChatInteractive:
             chat_completion,
             ignore_working,
             rawdog=rawdog,
-            external_exec=external_exec,
+            internal_exec=internal_exec,
             confirm_script=confirm_script,
             interpreter=interpreter,
         )
@@ -1239,7 +1338,7 @@ class ChatGenerate:
     """Generate command"""
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.option(
         "-m",
         "--model",
@@ -1276,8 +1375,7 @@ class ChatGenerate:
     @click.option(
         "-k",
         "--key",
-        envvar="auth_key",
-        help="LLM API access key or auth value",
+        help="LLM API access key or auth value or path to LLM with provider.",
     )
     @click.option(
         "-ct",
@@ -1355,7 +1453,6 @@ class ChatGenerate:
     @click.option(
         "-ho",
         "--history-offset",
-        envvar="history_offset",
         help="Limit conversation history to this number of last texts",
         type=click.IntRange(100, 16000),
         default=10250,
@@ -1373,7 +1470,6 @@ class ChatGenerate:
         "-pp",
         "--proxy-path",
         type=click.Path(exists=True),
-        envvar="proxy_path",
         help="Path to .json file containing proxies",
     )
     @click.option(
@@ -1382,7 +1478,6 @@ class ChatGenerate:
         type=click.Choice(pytgpt.available_providers),
         default=this.default_provider,
         help="Name of LLM provider.",
-        envvar="llm_provider",
         metavar=(
             f"[{', '.join(pytgpt.tgpt_providers)}] etc, "
             "run 'pytgpt gpt4free list providers -w' to view more providers"
@@ -1391,7 +1486,6 @@ class ChatGenerate:
     @click.option(
         "-vo",
         "--vertical-overflow",
-        envvar="vertical_overflow",
         help="Vertical overflow behaviour on content display",
         type=click.Choice(["visible", "crop", "ellipsis"]),
         default="ellipsis",
@@ -1400,7 +1494,7 @@ class ChatGenerate:
         "-q",
         "--quiet",
         is_flag=True,
-        help="Flag for controlling response-framing",
+        help="Flag for controlling response-framing and response verbosity",
         default=False,
     )
     @click.option(
@@ -1428,10 +1522,10 @@ class ChatGenerate:
         help="Generate and auto-execute Python scripts - (experimental)",
     )
     @click.option(
-        "-ex",
-        "--external-exec",
+        "-ix",
+        "--internal-exec",
         is_flag=True,
-        help="RawDog : Execute scripts with system's python interpreter",
+        help="RawDog : Execute scripts with exec function instead of out-of-script interpreter",
     )
     @click.option(
         "-cs",
@@ -1444,7 +1538,6 @@ class ChatGenerate:
         "--interpreter",
         default="python",
         help="RawDog : Python's interpreter name",
-        envvar="python_interpreter",
     )
     @click.help_option("-h", "--help")
     def generate(
@@ -1477,11 +1570,12 @@ class ChatGenerate:
         with_copied,
         ignore_working,
         rawdog,
-        external_exec,
+        internal_exec,
         confirm_script,
         interpreter,
     ):
         """Generate a quick response with AI"""
+        this.clear_history_file(filepath, new)
         bot = Main(
             max_tokens,
             temperature,
@@ -1501,7 +1595,7 @@ class ChatGenerate:
             quiet,
             ignore_working=ignore_working,
             rawdog=rawdog,
-            external_exec=external_exec,
+            internal_exec=internal_exec,
             confirm_script=confirm_script,
             interpreter=interpreter,
         )
@@ -1543,7 +1637,6 @@ class ChatGenerate:
             # {{stream}} without piped input
             raise Exception(f"No piped input detected ~ {stream_placeholder}")
 
-        this.clear_history_file(filepath, new)
         prompt = Optimizers.code(prompt) if code else prompt
         prompt = Optimizers.shell_command(prompt) if shell else prompt
         busy_bar.spin_index = (
@@ -1561,7 +1654,7 @@ class Awesome:
     """Awesome commands"""
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.option(
         "-r",
         "--remote",
@@ -1589,7 +1682,7 @@ class Awesome:
         )
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.argument(
         "key",
         required=True,
@@ -1632,7 +1725,7 @@ class Awesome:
             return resp != default
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.option("-n", "--name", required=True, help="Prompt name")
     @click.option("-p", "--prompt", required=True, help="Prompt value")
     @click.option(
@@ -1650,7 +1743,7 @@ class Awesome:
         return AwesomePrompts().add_prompt(name, prompt)
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.argument("name")
     @click.option(
         "--case-sensitive",
@@ -1674,7 +1767,7 @@ class Awesome:
         return AwesomePrompts().delete_prompt(name, case_sensitive)
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.option(
         "-j",
         "--json",
@@ -1729,7 +1822,7 @@ class Gpt4free:
     """Commands for gpt4free"""
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @busy_bar.run(index=1, immediate=True)
     @click.help_option("-h", "--help")
     def version():
@@ -1740,7 +1833,7 @@ class Gpt4free:
         click.secho(version_string, fg="cyan")
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.help_option("-h", "--help")
     @click.option(
         "-e",
@@ -1776,7 +1869,7 @@ class Gpt4free:
         click.secho(f"GPT4FREE updated successfully - {version_string}", fg="cyan")
 
     @staticmethod
-    @click.command("list")
+    @click.command("list", context_settings=this.context_settings)
     @click.argument("target")
     @click.option("-w", "--working", is_flag=True, help="Restrict to working providers")
     @click.option("-u", "--url", is_flag=True, help="Restrict to providers with url")
@@ -1829,6 +1922,7 @@ class Gpt4free:
                 if hunted_providers[0] is None
                 else hunted_providers
             )
+            hunted_providers.sort()
             if json:
                 rich.print_json(data=dict(providers=hunted_providers), indent=4)
 
@@ -1881,6 +1975,10 @@ class Gpt4free:
                 ],
             )
             if json:
+                for key, value in models.items():
+                    value.sort()
+                    models[key] = value
+
                 rich.print_json(data=models, indent=4)
             else:
                 table = Table(show_lines=True)
@@ -1888,13 +1986,13 @@ class Gpt4free:
                 table.add_column("Base Provider", style="cyan")
                 table.add_column("Model(s)", style="yellow")
                 for count, provider_models in enumerate(models.items()):
-                    table.add_row(
-                        str(count), provider_models[0], "\n".join(provider_models[1])
-                    )
+                    models = provider_models[1]
+                    models.sort()
+                    table.add_row(str(count), provider_models[0], "\n".join(models))
                 rich.print(table)
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.argument("port", type=click.INT, required=False)
     @click.option(
         "-a", "--address", help="Host on this particular address", default="127.0.0.1"
@@ -1923,12 +2021,78 @@ class Gpt4free:
             click.launch(f"http://{address}:{port}")
         t1.join()
 
+    @staticmethod
+    @click.command(context_settings=this.context_settings)
+    @click.option(
+        "-t",
+        "--timeout",
+        type=click.INT,
+        help="Provider response generation tiemout",
+        default=20,
+    )
+    @click.option(
+        "-r",
+        "--thread",
+        type=click.INT,
+        help="Test n amount of providers at once",
+        default=5,
+    )
+    @click.option("-q", "--quiet", is_flag=True, help="Suppress all stdout")
+    @click.option(
+        "-j", "--json", is_flag=True, help="Stdout test results in json format"
+    )
+    @click.option("-d", "--dry-test", is_flag=True, help="Return previous test results")
+    @click.option(
+        "-b", "--best", is_flag=True, help="Stdout the fastest provider <name only>"
+    )
+    @click.option("-y", "--yes", is_flag=True, help="Okay to all confirmations")
+    @click.help_option("-h", "--help")
+    def test(timeout, thread, quiet, json, dry_test, best, yes):
+        """Test and save working providers"""
+        from pytgpt.gpt4free import utils
+
+        test = utils.TestProviders(test_at_once=thread, quiet=quiet, timeout=timeout)
+        if best:
+            click.secho(test.best)
+            return
+        elif dry_test:
+            results = test.get_results(
+                run=False,
+            )
+        else:
+            if (
+                yes
+                or os.path.isfile(utils.results_path)
+                and click.confirm("Are you sure to run new test")
+            ):
+                results = test.get_results(run=True)
+            else:
+                results = test.get_results(
+                    run=False,
+                )
+        if json:
+            rich.print_json(data=dict(results=results))
+        else:
+            table = Table(
+                title="G4f Providers Test Results",
+                show_lines=True,
+            )
+            table.add_column("No.", style="white", justify="center")
+            table.add_column("Provider", style="yellow", justify="left")
+            table.add_column("Response Time(s)", style="cyan")
+
+            for no, provider in enumerate(results, start=1):
+                table.add_row(
+                    str(no), provider["name"], str(round(provider["time"], 2))
+                )
+            rich.print(table)
+
 
 class Utils:
     """Utilities command"""
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.argument("source", required=False)
     @click.option(
         "-d", "--dev", is_flag=True, help="Update from version control (development)"
@@ -1960,7 +2124,7 @@ class Utils:
         click.secho("Congratulations! Pytgpt updated successfully.", fg="cyan")
 
     @staticmethod
-    @click.command()
+    @click.command(context_settings=this.context_settings)
     @click.option("-w", "--whole", is_flag=True, help="Stdout whole json info")
     @click.option(
         "-v", "--version", is_flag=True, help="Stdout latest version name only"
@@ -2009,6 +2173,7 @@ def make_commands():
     EntryGroup.gpt4free.add_command(Gpt4free.update)
     EntryGroup.gpt4free.add_command(Gpt4free.show)
     EntryGroup.gpt4free.add_command(Gpt4free.gui)
+    EntryGroup.gpt4free.add_command(Gpt4free.test)
 
     # Awesome
     EntryGroup.awesome.add_command(Awesome.add)
